# frozen_string_literal: true

require 'import'
require 'cht-lib-fabric/tasks/actionable'
require_relative 'context_params_wrapper'

module AwsAccountStatus
  module Actions
    # This Action class does the work behind the "updateAccountRegions" API call.
    #
    # It saves the whole request to the data layer, and audits that a change was made.
    #
    class UpdateAccountRegions
      include ::Platform::Tasks::Actionable
      include Import['logger', 'statsd', 'data_access']

      # Update the enabled regions for an Account
      # See schema.graphql for input data definition
      def call
        update_start  = ::Time.now.to_f * 1000.0
        input_params  = context.params[:input]
        acct_id       = ContextParamsWrapper.new(input_params).entry_identifier
        input_regions = input_params[:regions]
        logger.debug "updating regions for  acct id: #{acct_id}, #regions: #{input_regions.length}"

        region_hash = input_regions.map do |input_row|
          [input_row[:name], { enabled: input_row[:enabled], enabled_by_default: input_row[:enabled_by_default] }]
        end.to_h

        begin
          data_access.update_region_status(acct_id, region_hash)

          # This is our audit log, only recorded on successful change, but record the whole state.
          input_regions.each do |r|
            name               = r[:name]
            isEnabled          = r[:enabled]
            isEnabledByDefault = r[:enabled_by_default]
            logger.info("UpdateRegions for account #{acct_id} - processing region name: #{name}, enabled: #{isEnabled}, enabledByDefault: #{isEnabledByDefault}")
          end
        ensure
          statsd.increment 'update_regions'
          update_end = ::Time.now.to_f * 1000.0
          elapsed    = update_end - update_start
          logger.debug "UpdateAccountRegions.call, elapsed time: #{elapsed} msec"
          statsd.timing 'update_acct_regions.duration', elapsed
          context['accountId'] = input_params[:account_id]
          context['customerId'] = input_params[:customer_id]
        end
        context['updateResult'] = 'success'
      end
    end
  end
end
