# frozen_string_literal: true

require 'redis'
require 'circuit_breaker'

module AwsAccountStatus
  module Clients
    # A Redis client with circuit breaker, only implementing the necessary commands for this service's
    # caching layer. The circuit breaker allows shortcutting Redis calls if Redis is regularly taking too
    # long to respond or the calls are producing consistent errors.
    #
    # +ttl+: most operations in this class take an optional +ttl+
    # parameter, defaulting to {DEFAULT_TTL}. If this value is
    # positive, then the expiration of the entire Redis key will be
    # updated to this many seconds in the future using an +EXPIRE+
    # call. If it is zero or negative, no change will be performed.
    # For composite values (lists, sets, hashes) this expiration
    # applies to the _entire_ key (list, set, hash), not individual
    # values. If you consistently call methods with +ttl=0+ then keys
    # will not expire.
    class Redis
      include CircuitBreaker

      DEFAULT_TTL = 1_814_400 # 3 weeks

      # Create a new Redis client.
      #
      # @param redis_client [Redis] optional. To provide an existing Redis client, or a mock.
      # @param options [Hash] passed on to {Redis} initializer
      def initialize(redis_client = nil, options = {})
        @redis_client = redis_client || ::Redis.new(options)
      end

      # Manually set the expiration of a key. This is a no-op if the requested +ttl+ is zero or negative.
      # This performs a Redis +EXPIRE+ operation.
      #
      # @param key [String] Redis key for the object
      # @param ttl [Integer] expiration, in seconds
      def expire(key, ttl)
        return if ttl <= 0

        call_redis :expire, key, ttl
      end

      # Add a value to a set, and update the expiration time of the set.
      # This performs a Redis +SADD+ operation.
      #
      # @param key [String] Redis key to update
      # @param value [String] value to add to the set
      # @param ttl [Integer] if positive, expire the key after this many seconds.
      def add_to_set(key, value, ttl = DEFAULT_TTL)
        call_redis(:sadd, key, value).tap do
          expire key, ttl
        end
      end

      # Get the members of a set. This performs a Redis +SMEMBERS+ operation.
      #
      # @param key [String] Redis key to get
      def get_set_members(key)
        call_redis(:smembers, key)
      end

      # Removes the members of a set. This performs a Redis +SREM+ operation.
      #
      # @param key [String] Redis key to remove
      # @param values [Array<String>] Values to remove from set
      def remove_set_members(key, values)
        call_redis(:srem, key, values)
      end

      # Delete an arbitrary key. This performs a Redis +DEL+ operation.
      #
      # @param key [String] Redis key to delete
      def clear(key)
        call_redis :del, key
      end

      private

      # Actually make the call to the base Redis client.
      # This method is protected by a circuit breaker. If it has many consecutive failures or timeouts,
      # the call will be skipped and instead raise a +::CircuitBreaker::CircuitBrokenException+.
      # After a delay, calls will attempt to go through to Redis again.
      def call_redis(func, *args)
        @redis_client.send(func, *args)
      end
      circuit_method :call_redis

      circuit_handler do |handler|
        # The number of failures needed to trip the breaker.
        handler.failure_threshold = 10

        # The period of time in seconds before attempting to reset the breaker.
        handler.failure_timeout = 30

        # How long in seconds can the call take before it's counted as a failure?
        handler.invocation_timeout = 1
      end
    end
  end
end
