require 'import'

module AwsAccountStatus
  module DataAccess
    # Provides access to the MySQL-implemented data store
    class AccountApiStatusSQLDataAccess
      include Import['logger']

      # Add/Update the API access status for a given customer / account / region / service / api
      # @param params_batch - An array of serialized Kafka messages. Each message contains
      #   all the information needed to create or update an AccountApiAccessStatus record.
      # @return [Boolean] indicating if an update took place
      def upsert_access_statuses(params_batch)
        status_hashes = create_status_hashes_array(params_batch)
        # Don't try to upsert if there's no data.
        return false if status_hashes.empty?

        options = Hash.new
        # mysql2 doesn't support (or need) :unique_by, but sqlite3
        # (used for running tests) requires it to work correctly.
        if ActiveRecord::Base.connection.adapter_name.include? 'SQLite'
          options[:unique_by] = :cust_id_acct_id_svc_reg_api_action_idx
        end

        begin
          AwsAccountStatus::Models::AccountApiAccessStatus
            .upsert_all(status_hashes, options)
          logger.info "Updated API access statuses: #{status_hashes}"
          true
        rescue => ex
          # If we hit an unexpected error, log it and keep going.
          logger.error { "Unable to save AccountApiAccessStatus record: #{ex}" }
          false
        end
      end

      # Get access statuses data for the given params
      # @param customer_id[Bigint] ID of customer
      # @param account_crn[String] AWS Account CRN
      # @param service[String] service name, for example: ec2, sagemaker
      # @param api_action[String] api action within the service, for example: describe-instances
      # @param region[String] AWS Region name
      # @return [Array<String>] List of accessible regions
      def get_access_statuses(customer_id:, account_crn:, service:, api_action:, region:)
        AwsAccountStatus::Models::AccountApiAccessStatus
          .where(
            customer_id: customer_id,
            account_crn: account_crn,
            service: service,
            api_action: api_action,
            region: region
          )
      end

      # Get the count of APIs with accessibility status for given customer
      # @param customer_id[Bigint] ID of customer
      # @param account_crns[Array<String>] (Optional) AWS account CRNs to scope to
      # @param accessible[Boolean] API accessibility status
      # @return [Integer] Count of number of APIs
      def get_api_count_with_status(customer_id:, account_crns:, accessible:)
        where_clause = {
          customer_id: customer_id,
          account_crn: account_crns,
          accessible: accessible
        }
        AwsAccountStatus::Models::AccountApiAccessStatus
          .where(where_clause)
          .count(:all)
      end

      # Get the count of APIs by account_crn with accessibility status for given customer
      # @param customer_id[Bigint] ID of customer
      # @param account_crn[Array<String>] List of account_crns to filter by
      # @return [Hash] Hash, [account_crn,accessible]=>count
      def get_account_crn_and_count(customer_id:, account_crn: nil)
        where_clause = {customer_id: customer_id}
        where_clause[:account_crn] = account_crn unless account_crn.nil?

        AwsAccountStatus::Models::AccountApiAccessStatus
          .group(:account_crn, :accessible)
          .where(where_clause)
          .count
      end

      # Get the API Access status for given customer/account(s)(optional)
      # @param customer_id[Bigint] ID of customer
      # @param account_crns[Array<String>] (Optional) AWS account CRNs to scope to
      # @return [Array<Hash>] API Access Status for each API
      def api_statuses(customer_id:, account_crns:)
        where_clause = {
          customer_id: customer_id,
          account_crn: account_crns
        }
        AwsAccountStatus::Models::AccountApiAccessStatus
          .where(where_clause)
          .map { |r| r.attributes.symbolize_keys }
      end

      private

      # Kafka message format:
      # {
      #   "parser"=>Karafka::Parsers::Json,
      #   "partition"=>0,
      #   "offset"=>2041,
      #   "key"=>"2017+3367254377911+kinesis+list_delivery_streams+us-east-1",
      #   "create_time"=>2021-06-24 11:27:06 +0000,
      #   "receive_time"=>2021-06-24 11:27:26 +0000,
      #   "topic"=>"com.cloudhealthtech.cp.collection.api_status",
      #   "parsed"=>true,
      #   "last_updated"=>"2021-06-24T11:27:06.771+00:00",
      #   "accessible"=>true
      # }

      def create_status_hashes_array(params_batch)
        statuses = []
        params_batch.each do |message|
          key = disassemble_key(message["key"])
          status = {
            customer_id: key[:customer_id],
            account_crn: key[:account_crn],
            service: key[:service],
            api_action: key[:api_action],
            region: key[:region],
            accessible: message["accessible"],
            last_checked: message["last_updated"]
          }
          # Ignore a status if any value is missing.
          statuses << status if status.values.all? { |v| !v.nil? }
        end
        statuses
      end

      def disassemble_key(key)
        Hash[[:customer_id, :account_crn, :service, :api_action, :region].zip(key.split('+'))]
      end
    end
  end
end
