require 'aws-sdk-dynamodb'
require "active_support/core_ext/hash/indifferent_access"
require "active_support/core_ext/time"

module AwsAccountStatus
  module DataAccess
    # Provides access to the DynamoDB-implemented data store
    class DynamoDb
      # table_name: required.  The name of the table to access
      # dynamo_db_client: optional.  If no client is provided, use the default AWS SDK client, otherwise use the
      #   provided one.  Useful for providing a mock client or specially configured client.
      def initialize(table_name, dynamo_db_client = nil)
        if dynamo_db_client
          @client = dynamo_db_client
        else
          @client = Aws::DynamoDB::Client.new
        end
        @table_name = table_name
      end

      # visible for testing wiring, only.
      attr_reader :client
      attr_reader :table_name

      # Update the Region Status for an AWS Account
      # account_id: the AWS account ID to update status for
      # region_status: the status to update to.  Will be in the form of a map of region name to status pair.  More
      #   specifically:
      #     {region_name: [enabled, enabled_by_default]}
      #   An example of specifying that us-east-1 is enabled and enabled by default, while us-west-2 is disabled, but
      #   enabled by default:
      #     {
      #       'us-east-1': { enabled: true, enabled_by_default: true },
      #       'us-west-2': { enabled: false, enabled_by_default: true }
      #     }
      # @return boolean indicating if an update took place
      def update_region_status(account_id, region_status)
        update_took_place = false

        # TODO, there might be a race condition in this method if >=2 clients are updating at the same time.
        # DynamoDB is only transactional per-row, so we need to know all the rows that exist before we change anything.
        before = internal_get_status_for_all_regions(account_id, true)
        # then we remove things that aren't in the incoming
        to_remove = before.keys - region_status.keys
        to_remove.each do |region_to_remove|
          update_took_place = true
          @client.delete_item({
                                  key:        {
                                      account_id: account_id,
                                      region:     region_to_remove,
                                  },
                                  table_name: @table_name,
                              })
        end

        # And store the incoming, one-by-one
        # ONLY when there are differences, or when the expiration_time needs to be updated
        region_status.each do |region, status_arr|
          diffs_exist =
            (before[region].nil?) || # It's a new record, OR
            # It's an existing record with changes, OR
            (status_arr != before[region].slice(:enabled, :enabled_by_default)) ||
            # It's near expiration
            (before[region][:expiration_time] < now.advance(days: 7).to_i)

          if diffs_exist
            update_took_place = true
            @client.put_item({
                                 item:       {
                                     account_id:         account_id,
                                     region:             region,
                                     enabled:            b_to_i(status_arr[:enabled]),
                                     enabled_by_default: b_to_i(status_arr[:enabled_by_default]),
                                     expiration_time:    now.advance(months: 1).to_i,
                                 },
                                 table_name: @table_name,
                             })
          end
        end

        update_took_place
      end

      # Gets the set of enabled regions for an AWS account ID.  Result is an array of region names, like:
      # [ 'us-east-1', 'us-west-2' ]
      # or an empty array, [], if there are no entries for the account
      def get_enabled_regions(account_id)
        enabled_regions = execute_query({
                                            table_name:                  @table_name,
                                            index_name:                  'account_enabled',
                                            key_condition_expression:    'account_id = :account_id and enabled = :enabled',
                                            expression_attribute_values: {
                                                ':account_id' => account_id,
                                                ':enabled'    => b_to_i(true)
                                            }
                                        })

        # Just get the region name.
        enabled_regions.keys
      end

      # Gets the current status for all known regions related to the specified AWS account.  Response is a map of
      # region name to status.  Example:
      #     {
      #       'us-east-1': { enabled: true, enabled_by_default: true },
      #       'us-west-2': { enabled: false, enabled_by_default: true }
      #     }
      def get_status_for_all_regions(account_id)
        internal_get_status_for_all_regions(account_id)
      end

      # Raises exception if there's an issue
      def health_check
        @client.describe_table({
                                   table_name: @table_name,
                               })
        nil
      end

      # Visible for override in testing.
      def now
        Time.now
      end

      private

      def internal_get_status_for_all_regions(account_id, include_expiration = false)
        execute_query({
                          table_name:                  @table_name,
                          key_condition_expression:    'account_id = :account_id',
                          expression_attribute_values: {
                              ':account_id' => account_id
                          }
                      },
                      include_expiration)
      end

      def execute_query(query, include_expiration = false)
        response = @client.query(query)
        response_to_hash(response.items, include_expiration)
      end

      def response_to_hash(items, include_expiration = false)
        #[{"region"=>"us-east-1", "enabled"=>0.1e1, "account_id"=>"1234", "enabled_by_default"=>true}]
        items.map do |i|
          if include_expiration
            [i["region"], { enabled: i_to_b(i["enabled"]), enabled_by_default: i_to_b(i["enabled_by_default"]), expiration_time: i["expiration_time"] }]
          else
            [i["region"], { enabled: i_to_b(i["enabled"]), enabled_by_default: i_to_b(i["enabled_by_default"]) }]
          end
        end.to_h
        #  { 'us-east-1': { enabled: true, enabled_by_default: true } }
      end

      ##
      # These conversions are required because DynamoDB doesn't like to keep BOOL types in indicies.  Instead, we use
      # numeric types, and just convert to bool here.
      ##

      # Converts bool to integer
      def b_to_i(bool)
        bool ? 1 : 0
      end

      # Converts integer to Bool
      def i_to_b(int)
        int == 1
      end
    end
  end
end
