# frozen_string_literal: true
require 'aws_account_status/data_access/dynamodb_data_access'

module AwsAccountStatus
  module DataAccess
    # A proxy hiding a DynamoDB Data Access object. This is helpful to allow the DynamoDB client to not be
    # initialized unless needed. In many API cases, Redis is all that's needed for the request, skipping
    # an AWS credentials call and the relatively-expensive initialization of the DynamoDB client.
    class DynamoDbDataAccessProxy
      def initialize(table_name)
        @dynamo_access = nil # Initialized at first call.
        @dynamo_table_name = table_name
      end

      # If anything is called against this class, initialize the DynamoDB layer if needed and pass the call
      # through.
      def method_missing(method_name, *args, &block)
        ensure_dynamo_setup

        if @dynamo_access.respond_to?(method_name)
          @dynamo_access.send(method_name, *args, &block)
        else
          super
        end
      end

      def respond_to_missing?(method_name, *args)
        ensure_dynamo_setup

        @dynamo_access.respond_to?(method_name, *args)
      end

      private

      def ensure_dynamo_setup
        return if @dynamo_access

        @dynamo_access = ::AwsAccountStatus::DataAccess::DynamoDb.new(
          @dynamo_table_name,
          Application.resolve('dynamodb.client') # Get the shared DynamoDB client.
        )
      end
    end
  end
end
