# frozen_string_literal: true

module AwsAccountStatus
  module DataAccess
    # Redis caching layer.
    class Redis
      NO_REGIONS_SENTINEL = 'no_regions'

      # Create a Redis Data Access object, backed by the provided other data access object. This allows
      # chaining the data access layers.
      #
      # @param redis_client [AwsAccountStatus::Clients::Redis] The internal Redis client for Redis
      #   communication. Useful for mocking out calls during testing.
      # @param backing_data_access [AwsAccountStatus::DataAccess::DynamoDbDataAccessProxy] Another +DataAccess+
      #   object to be covered by the cache. This will normally be the DynamoDB data access proxy layer.
      # @param logger [Logger] Logger to use.
      def initialize(redis_client, backing_data_access, logger)
        @redis = redis_client
        @backing_data_access = backing_data_access
        @logger = logger
      end

      # Because this is just a wrapper around the DynamoDB access, delegate unhandled methods back to
      # DynamoDB.
      delegate :get_status_for_all_regions, :health_check, to: :@backing_data_access

      # Update the Region Status for an AWS Account
      # account_id: the account ID to update status for.
      # region_status: the status to update to. Will be in the form of a map of region name to status pair.
      #   An example of specifying that us-east-1 is enabled and enabled by default, while us-west-2 is
      #   disabled, but enabled by default:
      #     {
      #       'us-east-1': { enabled: true, enabled_by_default: true },
      #       'us-west-2': { enabled: false, enabled_by_default: true }
      #     }
      # @return boolean indicating if an update took place
      def update_region_status(account_id, region_status)
        # Pass the update call straight through to the Dynamo layer
        update_took_place = @backing_data_access.update_region_status(account_id, region_status)

        # If nothing changed, we can return early.
        return false unless update_took_place

        begin
          # If there were changes, invalidate the cache for this account.
          @redis.clear(redis_key(account_id))
        rescue ::CircuitBreaker::CircuitBrokenException
          @logger.warn { "Skipping clearing Redis cache on update. Circuit breaker open." }
        end

        true
      end

      # Gets the set of enabled regions for an AWS account ID. Result is an array of region names, like:
      # [ 'us-east-1', 'us-west-2' ]
      # or an empty array, [], if there are no entries for the account.
      def get_enabled_regions(account_id)
        redis_accessible = true
        begin
          # Check the cache for the data.
          result = @redis.get_set_members(redis_key(account_id))
        rescue ::CircuitBreaker::CircuitBrokenException
          @logger.warn { "Skipping get from Redis. Circuit breaker open." }
          result = nil # Ensure it's +nil+ to allow a call to DynamoDB below.

          # If we failed to get data from Redis, don't try to update the cache after getting it from DynamoDB.
          # This avoids a potential problem where we could add duplicate/incorrect data to the cache.
          redis_accessible = false
        end

        # If Redis returned data, handle it and return.
        if result.present?
          # Check if the data in the cache was the sentinel.
          if result.include?(NO_REGIONS_SENTINEL)
            # If the only data cached was the indicator that the account has no regions, return no regions.
            return [] if result.length == 1

            # Somehow the sentinel was in the list with other data. This shouldn't happen.
            @logger.error { "no_regions sentinel mixed with valid data for account_id: #{account_id}" }
            # Remove the sentinel from the returned data.
            result.delete(NO_REGIONS_SENTINEL)
          end

          return result
        end

        # The data wasn't in the cache, so check DynamoDB.
        result = @backing_data_access.get_enabled_regions(account_id)

        # Skip writing to Redis if we failed to read from Redis initially.
        return result unless redis_accessible

        if result.present?
          # Cache the response in Redis
          create_redis_cache(account_id, result)
        else
          # No DynamoDB data, so cache the sentinel and return empty data.
          @logger.debug { 'No data returned from DynamoDB layer. Caching empty data marker.' }
          create_redis_cache(account_id, NO_REGIONS_SENTINEL)
        end

        result
      end

      private

      def create_redis_cache(account_id, data)
        @redis.add_to_set(redis_key(account_id), data)
      rescue ::CircuitBreaker::CircuitBrokenException
        @logger.warn { "Skipping update call to Redis. Circuit breaker open." }
      end

      # Get the key used for Redis storage based on the given account ID.
      def redis_key(account_id)
        "account:#{account_id}:enabled_regions"
      end
    end
  end
end
