# frozen_string_literal: true

module AwsAccountStatus
  module DataAccess

    class RegionServiceSupportRedisDataAccess
      TTL_NO_EXPIRY = -1

      # Create a Redis Data Access object for accessing RegionServiceSupport.
      #
      # @param redis_client [AwsAccountStatus::Clients::Redis] Internal Redis client for Redis communication. Useful for mocking out calls during testing.
      # @param logger [Logger] Logger to use.
      def initialize(redis_client, logger)
        @redis = redis_client
        @logger = logger
      end

      # Get supported regions from Redis for a service.
      #
      # @param service [String] Name of the service
      def get_supported_regions_by_service(service)
        @redis.get_set_members(redis_key(service))
      end

      # Insert supported services into Redis and remove any services which are not supported anymore for a region.
      #
      # @param region_service_support [AwsAccountStatus::Models::RegionServiceSupport] Region to update along with its supported or not supported services.
      def update_supported_region_by_service(region_service_support)
        region = region_service_support.region
        supported_services = region_service_support.services.select { |service| service.supported? }.map { |service| service.name }
        unsupported_services = region_service_support.services.select { |service| not service.supported? }.map { |service| service.name }

        remove_supported_region_by_services(region, unsupported_services)
        add_supported_regions_by_services(region, supported_services)
      end

      protected

      # Add services to the list of supported services for the region
      #
      # @param region [String] Region for the service
      # @param services [Array<String>] A list of services to be added
      def add_supported_regions_by_services(region, services)
        services.each do |service|
          @redis.add_to_set(redis_key(service), region, TTL_NO_EXPIRY)
        end
      end

      # Remove services from the list of supported services for the region
      #
      # @param region [String] Region for the service
      # @param services [Array<String>] A list of services to be removed
      def remove_supported_region_by_services(region, services)
        services.each do |service|
          @redis.remove_set_members(redis_key(service), region)
        end
      end

      # Get the redis key for a given service
      #
      # @param service [String] Name of the service
      def redis_key(service)
        "service:#{service}:supported_regions"
      end

    end

  end
end
