require 'hitimes'
require './app/aws_account_status/graph_ql/types/resolvers/base_api_access_resolver'

module AwsAccountStatus
  module GraphQL
    module Types
      module Resolvers
        class AwsAccountsApiAccessStatusResolver < BaseApiAccessResolver

          def call(args, context)
            duration_metric = Hitimes::TimedMetric.new('aws_accounts_api_status_resolver')
            duration_metric.start
            begin
              resolve(args, context)
            ensure
              duration_metric.stop
              record_metrics duration_metric, 'aws_accounts_api_status'
            end
          end

          private

          def resolve(args, context)
            nodes, has_previous_page, has_next_page = get_accounts_from_org_mgmt(args, context)
            populate_api_counts!(customer_id_from(context), nodes)
            start_cursor, end_cursor = [nodes[0][:cursor], nodes[-1][:cursor]] if nodes.size > 0
            filtered_nodes = get_filtered_result(nodes, args)

            ::AwsAccountStatus::GraphQL::Types::AwsAccountsApiAccessStatusResolverResult.new(
              nodes: filtered_nodes,
              has_previous_page: has_previous_page,
              has_next_page: has_next_page,
              start_cursor: start_cursor || '',
              end_cursor: end_cursor || ''
            )
          end

          private

          def get_filtered_result(nodes, args)
            filter_for_inaccessible_api(nodes, args)
          end

          def populate_api_counts!(customer_id, nodes)
            crn_nodes_lookup = nodes.map { |n| [n[:id], n] }.to_h
            api_statuses = api_status_data_access.api_statuses(
              customer_id: customer_id,
              account_crns: crn_nodes_lookup.keys
            )

            api_counts = get_api_counts(customer_id, api_statuses)

            api_counts.to_a.map { |a| a.flatten(1) }.each do |crn, is_accessible, count|
              keys_to_update = is_accessible ? :accessible_api_count : :inaccessible_api_count
              crn_nodes_lookup[crn][keys_to_update] = count
            end
          end

          def get_accounts_from_org_mgmt(args, context)
            is_authorized?(context)

            customer_id = customer_id_from(context)
            organization_id = organization_id_from(context)
            organization_crn = build_organization_crn(customer_id, organization_id)
            org_mgmt_res = get_accounts_in_org_page(organization_crn, context, args)
            nodes = org_mgmt_res["edges"].map do |edge|
              node = {}
              node[:cursor] = edge["cursor"]
              node[:name] = edge["node"]["name"]
              node[:id] = edge["node"]["id"]
              node[:status] = edge["node"]["status"]
              node[:inaccessible_api_count] = 0
              node[:accessible_api_count] = 0

              node
            end

            has_previous_page = org_mgmt_res["pageInfo"]["hasPreviousPage"]
            has_next_page = org_mgmt_res["pageInfo"]["hasNextPage"]

            [nodes, has_previous_page, has_next_page]
          end

          def filter_for_inaccessible_api(nodes, args)
            if args[:filter_rules].nil?
              return nodes
            end

            inaccessible_api_count_filter = args[:filter_rules].select { |f| f.field == "inaccessibleApiCount" }[0]
            if inaccessible_api_count_filter.nil? || inaccessible_api_count_filter.filter_type != "GT"
              return nodes
            end

            count = inaccessible_api_count_filter.filter_values[0].to_i

            nodes.select { |n| n[:inaccessible_api_count] > count }
          end

          # Get the count of APIs of api_statuses with accessibility status for given customer
          # Note: We are ignoring count of APIs that are part of spendless regions
          # @param customer_id[Bigint] ID of customer
          # @param api_statuses[Array<String>] List of api_status filter by customer and list of account_crn
          # @return [Hash] Hash, [account_crn,accessible]=>count
          def get_api_counts(customer_id, api_statuses)
            # Get spendless regions for all `account_crn` after filtering them
            spendless_regions = get_spendless_region_and_api(customer_id, api_statuses.pluck(:account_crn).uniq)

            # Reject account status that are part of spendless regions and ignore it from count of APIs
            api_statuses
              .reject { |as| spendless_regions["#{as[:account_crn]}+#{as[:service]}+#{as[:api_action]}"]&.include?(as[:region]) }
              .group_by { |as| [as[:account_crn], as[:accessible]] }
              .transform_values(&:size)
          end

          # Get the list of spendless regions for list of account_crn based on `account_crn`, `service`, and `api_name`
          # @param customer_id[Bigint] ID of customer
          # @param account_crns[Array<String>] List of account_crn
          # @return [Hash] Hash, ["account_crn+service+api"]=>["regions"]
          def get_spendless_region_and_api(customer_id, account_crns)
            account_crns.each_with_object({}) do |ac, hash|
              collection_status = get_collection_history(customer_id, ac)
              spendless_regions = get_spendless_region(customer_id, ac)

              spendless_regions.each do |sr|
                service, asset = sr["service_asset"].split("+")
                status = collection_status.find { |cs| cs[:service] == service && cs[:asset] == asset }
                next unless status && status[:required_iam_permissions]
                status[:required_iam_permissions].each { |s| hash["#{ac}+#{service}+#{s[:api]}"] = sr["regions"] }
              end
            end
          end
        end
      end
    end
  end
end
