require 'import'
require 'cht-lib-fabric/helpers/graphql/raise_error'
require 'cht-lib-fabric/error_codes'
require 'active_support/inflector'

module AwsAccountStatus
  module GraphQL
    module Types
      module Resolvers
        class BaseApiAccessResolver
          include Import["logger", "statsd", "api_status_data_access", "integration_aws",
                         "org_management", "authorization", "spend_data_access"]
          include ::Platform::ErrorCodes
          include ::Platform::Helpers::GraphQL::RaiseError

          private

          def build_organization_crn(customer_id, organization_id)
            "crn:#{customer_id}:organization/#{organization_id}"
          end

          def customer_id_from(context)
            context[:request]&.current_customer
          end

          def organization_id_from(context)
            context[:request]&.current_organization
          end

          def get_accounts_crns_in_org(organization_crn, context, account_crns: [])
            logger.debug "Getting list of accounts in organization from org_management"
            accounts = org_management.fetch_accounts_in_org(org_crn: organization_crn, context: context, account_crns: account_crns)
            accounts.inject([]) { |arr, val| arr << val[:id] }
          end

          def get_accounts_in_org_page(organization_crn, context, args)
            logger.debug "Getting list of accounts in organization from org_management"

            supported_args = [:first, :last, :before, :after]
            forwarding_args = args.dup.slice(*supported_args)

            unless args[:filter_rules].nil?
              filters_to_exclude = %w[inaccessibleApiCount accessibleApiCount]
              forwarding_filters = args[:filter_rules].
                dup.
                select { |f| !filters_to_exclude.include?(f.field) }.
                map { |f| f.to_h.transform_keys { |k| ActiveSupport::Inflector.camelize(k, uppercase_first_letter = false).to_s } }
              forwarding_args[:filters] = forwarding_filters if forwarding_filters.size > 0
            end

            unless args[:sort_rules].nil?
              forwarding_sort_rules = args[:sort_rules].dup.map(&:to_h)
              forwarding_args[:sort] = forwarding_sort_rules if forwarding_sort_rules.size > 0
            end

            org_management.fetch_accounts_in_org_page(org_crn: organization_crn, context: context, args: forwarding_args)
          end

          def get_api_status(customer_id, account_crns)
            logger.debug "Getting API Status from data access"
            api_status_data_access.api_statuses(customer_id: customer_id, account_crns: account_crns)
          end

          def get_collection_history(customer_id, account_crn)
            logger.debug "Getting collection_status from integration_aws"
            integration_aws.fetch_collection_status(customer_id: customer_id, account_crn: account_crn)
          end

          def get_spendless_region(customer_id, account_crn)
            logger.debug "Getting Spendless Region from spend data access"
            account_id = account_crn ? account_crn.split('/').last : nil
            spend_data_access.get_spendless_regions_by_customer_account(customer_id, account_id)
          end

          def get_is_authorized(context)
            logger.debug "Getting is_authorized from authorization"
            authorization.is_authorized(context: context)
          end

          def is_authorized?(context)
            unless get_is_authorized(context)
              throw raise_error(
                ::Platform::ErrorCodes::NOT_AUTHORIZED,
                uuid: context[:request].request_id
              )
            end
          end

          def record_metrics(duration_metric, call)
            elapsed = get_duration_in_ms(duration_metric)
            logger.debug "#{call}.call, elapsed time: #{elapsed} msec"
            statsd.increment "#{call}.count"
            statsd.timing "#{call}.duration", elapsed
          end

          def get_duration_in_ms(duration_metric)
            (duration_metric.duration * 1000).to_i
          end
        end
      end
    end
  end
end
