require "set"

module AwsAccountStatus
  module Helpers
    module ResponseBuilder
      class AwsAccountsWithApiAccessStatusResponseBuilder

        SERVICE_BLACKLIST = %w(
          fanout-collection
          generic_derived_asset_scanner
        )

        ASSET_BLACK_LIST = %w(
          aws_region_service_support
          db-query
          synthetic_derived_asset_scanner
        )

        def initialize
          @api_statuses = []
          @collection_status = []
          @spendless_regions = []
          @collection_status_lookup = nil
        end

        def collection_status(collection_status)
          @collection_status.concat(collection_status)
          self
        end

        def api_statuses(api_statuses)
          @api_statuses.concat(api_statuses)
          self
        end

        def spendless_regions(spendless_regions)
          @spendless_regions.concat(spendless_regions)
          self
        end

        def build
          prune_collection_status!
          build_collection_status_lookup
          services = build_by_service(@api_statuses)
          add_disabled_services_from_collection services

          { services: services }
        end

        private

        def build_by_service(api_statuses)
          api_statuses.group_by { |a| a[:service] }.map do |service_name, api_statuses_in_service|
            regions = build_by_region(service_name, api_statuses_in_service)

            disabled = false
            disabled_reason = ''
            # Only try to get this info if we have collection status data for this service.
            if @collection_status_lookup[service_name]
              disabled = @collection_status_lookup[service_name][:disabled]
              disabled_reason = @collection_status_lookup[service_name][:disabled_reason]
              @collection_status_lookup[service_name][:added_to_response] = true
            end

            AwsAccountApiAccessStatusService.new(
              name: service_name,
              regions: regions,
              disabled: disabled,
              disabled_reason: disabled_reason
            )
          end
        end

        def prune_collection_status!
          @collection_status = @collection_status
                                 .select { |cs| !SERVICE_BLACKLIST.include?(cs[:service]) }
                                 .select { |cs| !ASSET_BLACK_LIST.include?(cs[:asset]) }
        end

        def build_by_region(service_name, api_statuses)
          api_statuses.group_by { |a| a[:region] }.map do |region_name, api_statuses_in_region|
            api_statuses = build_api_statuses(service_name, region_name, api_statuses_in_region)

            # Skip building any region if all of its API Statuses is part of spendless region
            next if api_statuses == []

            AwsAccountApiAccessRegion.new(
              name: region_name,
              api_access_statuses: api_statuses
            )
          end.compact
        end

        def build_api_statuses(service_name, region_name, api_statuses)
          api_statuses.map do |api_status|
            api_name = api_status[:api_action]
            assets = @collection_status_lookup.dig(service_name, :apis, api_name, :assets) || []
            service_assets = assets.map { |asset| "#{service_name}+#{asset}" }

            # Skip building any API Status that is part of spendless region
            next if @spendless_regions.any? { |sr| service_assets.include?(sr["service_asset"]) && sr["regions"].include?(region_name) }

            AwsApiAccessStatus.new(
              accessible: api_status[:accessible],
              assets: assets,
              api_name: api_status[:api_action],
              last_updated: api_status[:last_checked].to_s,
              recommendation: @collection_status_lookup.dig(service_name, :apis, api_name, :recommendation)
            )
          end.compact
        end

        def build_collection_status_lookup
          services = {}
          @collection_status.group_by { |c| c[:service] }.each do |service_name, collectors|
            disabled_reasons = Set.new
            disabled = false
            apis = {}
            collectors.each do |c|
              disabled ||= c[:status] == "disabled"
              disabled_reasons << c[:disabled_reason]
              c[:required_iam_permissions].each do |required_api|
                api_name = required_api[:api]
                api = apis[api_name]
                if api.nil?
                  api = required_api
                  api[:assets] = [c[:asset]]
                else
                  api[:assets] << c[:asset]
                end
                apis[api_name] = api
              end
            end
            disabled_reason = disabled_reasons.select { |d| d.present? }.to_a.join(",")
            services[service_name] = { :disabled => disabled, :disabled_reason => disabled_reason, :apis => apis }
          end

          @collection_status_lookup = services
        end

        def add_disabled_services_from_collection(services_response)
          disabled_services = @collection_status_lookup.select { |_, v| v[:added_to_response].nil? }.map do |service_name, service|
            AwsAccountApiAccessStatusService.new(
              name: service_name,
              regions: [],
              disabled: service[:disabled],
              disabled_reason: service[:disabled_reason]
            )
          end

          services_response.concat(disabled_services)
        end
      end

      class AwsApiAccessStatus < ::Dry::Struct
        attribute :api_name, ::Platform::DryTypes::Strict::String
        attribute :accessible, ::Platform::DryTypes::Strict::Bool
        attribute :assets, ::Platform::DryTypes::Strict::Array.of(::Platform::DryTypes::String)
        attribute :recommendation, ::Platform::DryTypes::String.optional
        attribute :last_updated, ::Platform::DryTypes::Strict::String
      end

      class AwsAccountApiAccessRegion < ::Dry::Struct
        attribute :name, ::Platform::DryTypes::Strict::String
        attribute :api_access_statuses, ::Platform::DryTypes::Strict::Array.of(AwsApiAccessStatus)
      end

      class AwsAccountApiAccessStatusService < ::Dry::Struct
        attribute :name, ::Platform::DryTypes::Strict::String
        attribute :disabled, ::Platform::DryTypes::Bool.optional
        attribute :disabled_reason, ::Platform::DryTypes::String.optional
        attribute :regions, ::Platform::DryTypes::Strict::Array.of(AwsAccountApiAccessRegion)
      end

      class AwsAccountApiAccessStatus < ::Dry::Struct
        attribute :services, ::Platform::DryTypes::Strict::Array.of(AwsAccountApiAccessStatusService)
      end
    end
  end
end
