require 'aws-sdk-dynamodb'
require "active_support/core_ext/hash/indifferent_access"
require "active_support/core_ext/time"

module AwsAccountStatus
  module SpendDataAccess
    # Provides access to the DynamoDB-implemented data store
    class DynamoDb
      # table_name: required.  The name of the table to access
      # dynamo_db_client: optional.  If no client is provided, use the default AWS SDK client, otherwise use the
      #   provided one.  Useful for providing a mock client or specially configured client.
      def initialize(table_name, dynamo_db_client = nil)
        if dynamo_db_client
          @client = dynamo_db_client
        else
          @client = Aws::DynamoDB::Client.new
        end
        @table_name = table_name
      end

      # visible for testing wiring, only.
      attr_reader :client
      attr_reader :table_name


      # Gets the set of spendless regions for an AWS customer/account. Result is an array of hashes,
      # like: {"regions"=>["us-east-1", "us-west-2"], "service_asset"=>"ec2+instance"}
      # or an empty array, [], if there are no entries for the index.
      # customer_id: String!
      # account_id: String!
      def get_spendless_regions_by_customer_account(customer_id, account_id)
        primary_key = "#{customer_id}+#{account_id}"

        spendless_regions = @client.query(
                                            table_name:                  @table_name,
                                            key_condition_expression:    'customer_id_account_id = :customer_id_account_id',
                                            expression_attribute_values: {
                                              ':customer_id_account_id' => primary_key
                                            })

        return [] if spendless_regions.count == 0
        # Get all spendless regions (by `regions` and `service_asset`) for an account
        spendless_regions.items.map { |sr| sr.slice("regions", "service_asset") }
      end

      # Gets the set of spendless regions for an AWS customer/account/service/asset. Result is an array of
      # region names, like: [ 'us-east-1', 'us-west-2' ] or an empty array, [], if there are no entries for the
      # index. The array stores all regions without historical spend
      # customer_id: String!
      # account_id: String!
      # service: String!
      # asset: String!
      def get_spendless_regions(customer_id, account_id, service, asset)
        primary_key = "#{customer_id}+#{account_id}"
        range_key = "#{service}+#{asset}"

        spendless_regions = @client.query(
                                            table_name:                  @table_name,
                                            key_condition_expression:    'customer_id_account_id = :customer_id_account_id and service_asset = :service_asset',
                                            expression_attribute_values: {
                                              ':customer_id_account_id' => primary_key,
                                              ':service_asset'          => range_key
                                            })

        return [] if spendless_regions.count == 0
        # Just get the region names.
        spendless_regions.items[0]["regions"]
      end

      # Gets the set of spendless regions for an AWS customer/account/service/asset. Result is the full json
      # response for a get spendless regions query.
      # customer_id: String!
      # account_id: String!
      # service: String!
      # asset: String!
      def get_spendless_regions_response(customer_id, account_id, service, asset)
        primary_key = "#{customer_id}+#{account_id}"
        range_key = "#{service}+#{asset}"

        spendless_regions = @client.query(
          table_name:                  @table_name,
          key_condition_expression:    'customer_id_account_id = :customer_id_account_id and service_asset = :service_asset',
          expression_attribute_values: {
            ':customer_id_account_id' => primary_key,
            ':service_asset'          => range_key
          })

        # Just get the region names.
        spendless_regions.items[0]
      end

      # Update the list of spendless regions for an AWS customer/account/service/asset.
      # customer_id: String!
      # account_id: String!
      # service: String!
      # asset: String!
      # spendless_regions: [String!]!
      def update_spendless_regions(customer_id, account_id, service, asset, spendless_regions)
        update_took_place = false
        # check to see whether list of spendless regions has changed before writing
        before = get_spendless_regions_response(customer_id, account_id, service, asset)

        diffs_exist =
          # It's a new record, OR
          (before.nil?) ||
          # It's an existing record without regions data, OR
          (before["regions"].nil?) ||
          # It's an existing record with changes, OR
          (before["regions"] != spendless_regions) ||
          # It's an existing record without expiration_time, OR
          (before["expiration_time"].nil?) ||
          # It's within a week of expiration
          (before["expiration_time"] < now.advance(days: 7).to_i)

        if diffs_exist
          update_took_place = true
          primary_key = "#{customer_id}+#{account_id}"
          range_key = "#{service}+#{asset}"

          @client.put_item(item:       {
                             customer_id_account_id:    primary_key,
                             service_asset:             range_key,
                             expiration_time:           generate_expiration_epoch,
                             regions:                   spendless_regions
                           },
                           table_name: @table_name,)
        end
        # return whether or not update took place
        update_took_place
      end

      # Visible for override in testing.
      def now
        Time.now
      end

      # Generate time until data should expire
      def generate_expiration_epoch
        now.advance(months: 1).to_i
      end

      # Raises exception if there's an issue
      def health_check
        @client.describe_table({
                                 table_name: @table_name,
                               })
        nil
      end
    end
  end
end

